import copy
import json
from enum import Enum


class LambdaInvocationType(Enum):
    """
    REF: Supported Lambda Event Types
    https://docs.aws.amazon.com/lambda/latest/dg/API_Invoke.html#API_Invoke_RequestSyntax
    """
    EVENT = 'Event'
    REQUEST_RESPONSE = 'RequestResponse'
    DRY_RUN = 'DryRun'


def transpose_fn_url_event(event: dict):
    """See Request event payload schema - Amazon API Gateway payload format version 2.0
    https://docs.aws.amazon.com/lambda/latest/dg/urls-invocation.html
    https://docs.aws.amazon.com/apigateway/latest/developerguide/http-api-develop-integrations-lambda.html

    :param event:
    :return: transposed Lambda-native event dict or passthrough event as-is
    """
    if not all(k in event.keys() for k in ("version", "routeKey", "headers", "requestContext", "isBase64Encoded")):
        # don't seem to be HTTP request, passing it through
        return event

    http_method = event['requestContext']['http']['method']

    # If request parameter and POST body having the same payload key then we merge them
    # and, POST body payload value take precedence
    payload = {}

    # transpose GET request parameter as Lambda event payload
    if "queryStringParameters" in event.keys():
        payload.update(event['queryStringParameters'])

    # transpose POST request body as Lambda event payload
    if http_method == "POST" and "body" in event.keys():
        body = event['body']
        content_type = event['headers']['content-type']
        # is_base64_encoded = event['isBase64Encoded']

        if "json" in content_type:
            if isinstance(body, str):
                payload.update(json.loads(body))
            else:
                payload.update(body)
        else:
            # If the content type of the request is binary, the body is base64-encoded.
            # Passthrough for now
            return event

    if not payload:
        # We can't seem to transpose the event, passthrough for now
        return event

    return copy.deepcopy(payload)



import logging
import uuid
from typing import List

from libumccr import libjson
from libumccr.aws import eb_client

logger = logging.getLogger(__name__)

MAX_BATCH_SIZE = 10


def get_event_buses(**kwargs):
    """Proxy to
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/events/client/list_event_buses.html

    :param kwargs:
    :return:
    """
    return eb_client().list_event_buses(**kwargs)


def emit_events(entries: List[dict]) -> dict:
    """Proxy to
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/events/client/put_events.html

    :param entries:
    :return:
    """
    return eb_client().put_events(Entries=entries)


def emit_event(entry: dict) -> dict:
    return emit_events([entry])


def dispatch_events(entries: List[dict], batch_size=10):
    """
    PutEvents has maximum number of 10 entries per API call. So, this function batch them up if more than 10 entries.
    https://docs.aws.amazon.com/eventbridge/latest/APIReference/API_PutEvents.html

    :param entries:
    :param batch_size:
    :return:
    """
    _batch_size = batch_size if batch_size < MAX_BATCH_SIZE else MAX_BATCH_SIZE
    _chunks = [entries[x:x + _batch_size] for x in range(0, len(entries), _batch_size)]
    responses = {}
    for chunk in _chunks:
        chunk_id = str(uuid.uuid4())
        resp = emit_events(chunk)
        responses[chunk_id] = resp

    logger.info(f"EVENT BUS RESPONSE: \n{libjson.dumps(responses)}")
    return responses



# -*- coding: utf-8 -*-
"""libsm module

Module interface with AWS Secrets Manager and, LRU cache for hit less to AWS endpoint if any
"""
import base64
import logging

from cachetools.func import lru_cache

from libumccr.aws import sm_client

logger = logging.getLogger(__name__)


@lru_cache(maxsize=64)
def get_secret(secret_name):
    """
    If you are in long-running process and querying rotating value from secret manager such as JWT token, you may
    clear the cache before get secret call. e.g.

        libumccr.aws.libsm.get_secret.cache_clear()

    See https://github.com/umccr/gridss-purple-linx-nf/pull/25/files

    :param secret_name:
    :return:
    """
    resp = sm_client().get_secret_value(
        SecretId=secret_name
    )

    if 'SecretString' in resp:
        return resp['SecretString']
    else:
        return base64.b64decode(resp['SecretBinary'])



import csv
import io
import logging
from unittest import TestCase, skip

from libumccr import libgdrive

from libumccr.aws import libssm

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class LibGDriveUnitTests(TestCase):
    # TODO https://github.com/umccr/libumccr/issues/5
    pass


class LibGDriveIntegrationTests(TestCase):
    # integration test hit actual File or API endpoint, thus, manual run in most cases
    # required appropriate access mechanism setup such as active aws login session
    # uncomment @skip and hit the each test case!
    # and keep decorated @skip after tested

    @skip
    def test_download_sheet1_csv(self):
        """
        python -m unittest tests.test_libgdrive.LibGDriveIntegrationTests.test_download_sheet1_csv
        """
        SHEET_ID = "/umccr/google/drive/lims_sheet_id"
        GDRIVE_SERVICE_ACCOUNT = "/umccr/google/drive/lims_service_account_json"
        lims_sheet_id = libssm.get_secret(SHEET_ID)
        account_info = libssm.get_secret(GDRIVE_SERVICE_ACCOUNT)

        bytes_data = libgdrive.download_sheet1_csv(account_info, lims_sheet_id)
        self.assertIsInstance(bytes_data, bytes)

        # try parse and print last few rows
        csv_reader = csv.DictReader(io.TextIOWrapper(io.BytesIO(bytes_data)))
        for row_number, row in enumerate(csv_reader):
            if row_number > 4116:
                print(row_number, row)



import os
from datetime import datetime
from unittest import TestCase

from mockito import when, mock, verify, unstub

from libumccr import libslack


class LibSlackUnitTests(TestCase):

    def setUp(self) -> None:
        os.environ['SLACK_CHANNEL'] = "#mock"

    def tearDown(self) -> None:
        del os.environ['SLACK_CHANNEL']
        unstub()

    def test_call_slack_webhook(self):
        """
        python -m unittest tests.test_libslack.LibSlackUnitTests.test_call_slack_webhook
        """
        mock_sender = "mock sender"
        mock_topic = "mock topic"
        mock_attachments = [
            {
                "title": "mock attachement",
                "text": "test",
            },
            {
                "title": "datetime",
                "text": str(datetime.now()),
            },
        ]
        mock_response = mock(libslack.http.client.HTTPResponse)
        mock_response.status = 200

        when(libslack.libssm).get_ssm_param(...).thenReturn("mock_webhook_id_123")
        when(libslack.http.client.HTTPSConnection).request(...).thenReturn('ok')
        when(libslack.http.client.HTTPSConnection).getresponse(...).thenReturn(mock_response)

        status = libslack.call_slack_webhook(mock_sender, mock_topic, mock_attachments)
        self.assertEqual(200, status)
        verify(libslack.libssm, times=1).get_ssm_param(...)



import logging
from datetime import datetime
from unittest import TestCase

from libumccr import libjson

# FIXME centralise test case logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class LibJsonUnitTests(TestCase):

    def test_loads_str(self):
        """
        python -m unittest tests.test_libjson.LibJsonUnitTests.test_loads_str
        """
        json_str = """{"json": "this is json string"}"""
        self.assertIsInstance(json_str, str)
        result = libjson.loads(json_str)
        logger.info(result)
        self.assertIsInstance(result, dict)

    def test_loads_bytes(self):
        """
        python -m unittest tests.test_libjson.LibJsonUnitTests.test_loads_bytes
        """
        json_bytes = b'[{"sample":"SBJ123456","SNV8":0.079}]\n'
        self.assertIsInstance(json_bytes, bytes)
        result = libjson.loads(json_bytes)
        logger.info(result)
        self.assertIsInstance(result, list)

    def test_dumps_fallback(self):
        """
        python -m unittest tests.test_libjson.LibJsonUnitTests.test_dumps_fallback
        """
        result = libjson.dumps({
            'date': datetime.now(),
            'id': 1,
            'name': "Some string",
            'bool': True,
        })
        logger.info(result)
        self.assertIsInstance(result, str)


class LibJsonIntegrationTests(TestCase):
    pass



import re

"""
SAMPLE SHEET REGEXES
"""

EXPERIMENT_REGEX_STR = {
    "top_up": r"(?:_topup\d?)",
    "rerun": r"(?:_rerun\d?)"
}

SAMPLE_ID_REGEX_STR = {
    "sample_id_non_control": r"(?:PRJ|CCR|MDX|TGX)\d{6}",
    "sample_id_control": r"(?:NTC|PTC)_\w+"
}

SAMPLE_ID_REGEX_STR["sample_id"] = r"(?:(?:{})|(?:{}))".format(
    SAMPLE_ID_REGEX_STR["sample_id_non_control"],
    SAMPLE_ID_REGEX_STR["sample_id_control"]
)

LIBRARY_REGEX_STR = {
    "id_int": r"L\d{7}",
    "id_ext": r"L{}".format(SAMPLE_ID_REGEX_STR["sample_id"]),
    "year": r"(?:L|LPRJ)(\d{2})\d+"
}

LIBRARY_REGEX_STR["id"] = r"(?:{}|{})(?:{}|{})?".format(
    LIBRARY_REGEX_STR["id_int"],
    LIBRARY_REGEX_STR["id_ext"],
    EXPERIMENT_REGEX_STR["top_up"],                             # TODO - could a top_up/rerun exist?
    EXPERIMENT_REGEX_STR["rerun"]
)

SAMPLE_REGEX_OBJS = {
    # Sample ID: https://regex101.com/r/Z7fvHt/1
    "sample_id": re.compile(SAMPLE_ID_REGEX_STR["sample_id"]),
    # https://regex101.com/r/Z7fvHt/2
    "library_id": re.compile(LIBRARY_REGEX_STR["id"]),
    # https://regex101.com/r/Yf2t8E/2
    "unique_id_full_match": re.compile("{}_{}".format(SAMPLE_ID_REGEX_STR["sample_id"], LIBRARY_REGEX_STR["id"])),
    # https://regex101.com/r/Yf2t8E/3
    # Use brackets to capture the sample id and the library id
    "unique_id": re.compile("({})_({})".format(SAMPLE_ID_REGEX_STR["sample_id"], LIBRARY_REGEX_STR["id"])),
    # https://regex101.com/r/pkqI1n/1
    "topup": re.compile(EXPERIMENT_REGEX_STR["top_up"]),
    # TODO - add regex 101
    "rerun": re.compile(EXPERIMENT_REGEX_STR["rerun"]),
    # https://regex101.com/r/nNPwQu/1
    "year": re.compile(LIBRARY_REGEX_STR["year"])
}



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

# Note:
#   Version scheme follow PEP440 https://peps.python.org/pep-0440/
#   _NOT_ SemVer https://semver.org

setup(
    name="libumccr",
    version="0.4.0rc1",
    author="UMCCR and Contributors",
    author_email="services@umccr.org",
    description="UMCCR Reusable Python modules",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/umccr/libumccr",
    license="MIT",
    packages=find_packages(exclude=("tests**", "docs")),
    project_urls={
        "Bug Tracker": "https://github.com/umccr/libumccr/issues",
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    extras_require={
        "dev": [
            "pipdeptree",
            "twine",
            "setuptools",
            "wheel",
            "build",
            "pdoc3",
            "pre-commit",
            "detect-secrets",
        ],
        "test": [
            "pytest",
            "pytest-cov",
            "flake8",
            "mockito",
            "tox",
            "nose2",
            "awscli-local",
        ],
        "all": [
            "Django",
        ],
    },
    install_requires=[
        "boto3",
        "botocore",
        "cachetools",
        "requests",
        "pandas",
        "gspread",
        "gspread-pandas",
        "google-auth",
    ],
)
